import datetime
import enum
import logging
from typing import Final

import pydantic

logger = logging.getLogger(__name__)


class ERDDAPProtocol(str, enum.Enum):
    TABLEDAP: Final = "tabledap"
    GRIDDAP: Final = "griddap"


class BBox(pydantic.BaseModel):
    lon_min: float
    lon_max: float
    lat_min: float = pydantic.Field(-90, ge=-90, le=90)
    lat_max: float = pydantic.Field(90, ge=-90, le=90)

    class Config:
        extra = "forbid"


class SymmetricBBox(BBox):
    lon_min: float = pydantic.Field(-180, ge=-180, le=180)
    lon_max: float = pydantic.Field(180, ge=-180, le=180)


class AsymmetricBBox(BBox):
    lon_min: float = pydantic.Field(0, ge=0, le=360)
    lon_max: float = pydantic.Field(360, ge=0, le=360)


class Constraints(pydantic.BaseModel):
    bbox: BBox = pydantic.Field(...)
    start_date: datetime.datetime = pydantic.Field(...)
    end_date: datetime.datetime = pydantic.Field(datetime.datetime.now())


class SymmetricConstraints(Constraints):
    bbox: SymmetricBBox = SymmetricBBox()


class AsymmetricConstraints(Constraints):
    bbox: AsymmetricBBox = AsymmetricBBox()


class ERDDAPDataset(pydantic.BaseModel):
    server_url: pydantic.HttpUrl
    protocol: ERDDAPProtocol = pydantic.Field(ERDDAPProtocol.TABLEDAP)
    dataset_id: str
    is_longitude_symmetric: bool



import random
import time
from typing import Type

import limits


def wait(wait_time: float = 0.1, jitter: bool = True) -> None:
    """Wait for `wait_time` + some random `jitter`"""
    if jitter:
        # Set jitter to be <= 1% of wait_time
        jitter_time = random.random() / (1 / wait_time * 100)
    else:
        jitter_time = 0
    time.sleep(wait_time + jitter_time)


class RateLimit:
    def __init__(
        self,
        rate_limit: limits.RateLimitItem = limits.parse("5/second"),
        storage: Type[limits.storage.Storage] = limits.storage.MemoryStorage,
        strategy: Type[limits.strategies.RateLimiter] = limits.strategies.MovingWindowRateLimiter,
    ) -> None:
        self.rate_limit = rate_limit
        self.storage = storage()
        self.strategy = strategy(self.storage)

    def reached(self, identifier: str, cost: int = 1) -> bool:
        return not self.strategy.hit(
            self.rate_limit,
            identifier,
        )



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import TypeVar

import numpy.typing as npt


StrDict = Dict[str, Any]
ScalarOrArray = TypeVar("ScalarOrArray", int, float, npt.NDArray[Any])



import datetime
import io
import logging
from typing import cast
from typing import Optional

import erddapy
import pandas as pd
import requests

from searvey.custom_types import StrDict
from searvey.models import Constraints
from searvey.models import ERDDAPDataset


logger = logging.getLogger(__name__)

DEFAULT_SEARVEY_SESSION = requests.Session()


def ts_to_erddap(ts: datetime.datetime) -> str:
    return ts.strftime("%Y/%m/%dT%H:%M:%SZ")


# Adapted from:
# https://github.com/ioos/erddapy/blob/524783242c8b973ffd9bf5ab9f20f70516752f07/erddapy/url_handling.py#L14-L33
# @functools.lru_cache(maxsize=256)
def urlopen(
    url: str,
    session: requests.Session,
    requests_kwargs: StrDict,
    timeout: float = 10,
) -> io.BytesIO:
    # logger.debug("Making a GET request to: %s", url)
    response = session.get(url, allow_redirects=True, timeout=timeout, **requests_kwargs)
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as err:
        raise requests.exceptions.HTTPError(f"{response.content.decode()}") from err
    data = io.BytesIO(response.content)
    data.seek(0)
    return data


def get_erddap_url(
    dataset: ERDDAPDataset,
    constraints: Constraints,
    response: str = "csvp",
) -> str:
    erddap = erddapy.ERDDAP(
        server=dataset.server_url,
        protocol=dataset.protocol.value,
        response=response,
    )
    erddap.dataset_id = dataset.dataset_id
    erddap.constraints = {
        "time>=": ts_to_erddap(constraints.start_date),
        "time<=": ts_to_erddap(constraints.end_date),
        "latitude>=": constraints.bbox.lat_min,
        "latitude<=": constraints.bbox.lat_max,
        "longitude>=": constraints.bbox.lon_min,
        "longitude<=": constraints.bbox.lon_max,
    }
    url = erddap.get_download_url()
    return cast(str, url)


def query_erddap(
    dataset: ERDDAPDataset,
    constraints: Constraints,
    session: requests.Session = DEFAULT_SEARVEY_SESSION,
    timeout: int = 10,
    requests_kwargs: Optional[StrDict] = None,
    read_csv_kwargs: Optional[StrDict] = None,
) -> pd.DataFrame:
    if requests_kwargs is None:
        requests_kwargs = {}
    if read_csv_kwargs is None:
        read_csv_kwargs = {}
    # Make the query and parse the result
    url = get_erddap_url(dataset=dataset, constraints=constraints)
    logger.debug(url)
    data = urlopen(url, session=session, requests_kwargs=requests_kwargs, timeout=timeout)
    df = pd.read_csv(data, **read_csv_kwargs)
    return df



from __future__ import annotations

from searvey.coops import get_coops_stations
from searvey.ioc import get_ioc_data
from searvey.ioc import get_ioc_stations
from searvey.stations import get_stations
from searvey.stations import Provider


__all__: list[str] = [
    "get_coops_stations",
    "get_ioc_data",
    "get_ioc_stations",
    "get_stations",
    "Provider",
]



"""
interface with the U.S. National Oceanic and Atmospheric Administration (NOAA) Center for Operational Oceanographic Products and Services (CO-OPS) API
https://api.tidesandcurrents.noaa.gov/api/prod/
"""
import json
import logging
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from enum import Enum
from functools import lru_cache
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Union

import geopandas
import numpy
import pandas
import requests
import shapely
import typepigeon
import xarray
from bs4 import BeautifulSoup
from bs4 import element
from geopandas import GeoDataFrame
from pandas import DataFrame
from pandas import Series
from shapely.geometry import MultiPolygon
from shapely.geometry import Point
from shapely.geometry import Polygon
from xarray import Dataset


logger = logging.getLogger(__name__)


class StationDataProduct(Enum):
    pass


class StationDataInterval(Enum):
    pass


class StationDatum(Enum):
    pass


class StationStatus(Enum):
    ACTIVE = "active"
    DISCONTINUED = "discontinued"


class Station(ABC):
    """
    abstraction of a specific data station
    """

    id: str
    location: Point

    def __init__(self, id: str, location: Point):
        self.id = id
        self.location = location

    @abstractmethod
    def product(
        self,
        product: StationDataProduct,
        start_date: datetime,
        end_date: datetime = None,
        interval: StationDataInterval = None,
        datum: StationDatum = None,
    ) -> Dataset:
        """
        retrieve data for the current station within the specified parameters

        :param product: name of data product
        :param start_date: start date
        :param end_date: end date
        :param interval: time interval of data
        :param datum: vertical datum
        :return: data for the current station within the specified parameters
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        return f'{self.__class__.__name__} - "{self.id}" {self.location}'


class StationQuery(ABC):
    """
    abstraction of an individual station data query
    """

    station_id: str
    product: StationDataProduct
    start_date: datetime
    end_date: datetime
    interval: StationDataInterval
    datum: StationDatum

    def __init__(
        self,
        station_id: str,
        product: StationDataProduct,
        start_date: datetime,
        end_date: datetime = None,
        interval: StationDataInterval = None,
        datum: StationDatum = None,
    ):
        self.station_id = station_id
        self.product = product
        self.start_date = start_date
        self.end_date = end_date
        self.interval = interval
        self.datum = datum

    @property
    @abstractmethod
    def query(self) -> Dict[str, Any]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def data(self) -> DataFrame:
        """
        :return: data for the current query parameters
        """
        raise NotImplementedError()

    def __str__(self) -> str:
        return f'{self.__class__.__name__} - {self.product} at station "{self.station_id}" between {self.start_date} and {self.end_date} over {self.interval} in {self.datum}'


class COOPS_Product(StationDataProduct):  # noqa: N801
    WATER_LEVEL = (
        "water_level"
        # Preliminary or verified water levels, depending on availability.
    )
    AIR_TEMPERATURE = "air_temperature"  # Air temperature as measured at the station.
    WATER_TEMPERATURE = "water_temperature"  # Water temperature as measured at the station.
    WIND = "wind"  # Wind speed, direction, and gusts as measured at the station.
    AIR_PRESSURE = "air_pressure"  # Barometric pressure as measured at the station.
    AIR_GAP = "air_gap"  # Air Gap (distance between a bridge and the water's surface) at the station.
    CONDUCTIVITY = "conductivity"  # The water's conductivity as measured at the station.
    VISIBILITY = (
        "visibility"  # Visibility from the station's visibility sensor. A measure of atmospheric clarity.
    )
    HUMIDITY = "humidity"  # Relative humidity as measured at the station.
    SALINITY = "salinity"  # Salinity and specific gravity data for the station.
    HOURLY_HEIGHT = "hourly_height"  # Verified hourly height water level data for the station.
    HIGH_LOW = "high_low"  # Verified high/low water level data for the station.
    DAILY_MEAN = "daily_mean"  # Verified daily mean water level data for the station.
    MONTHLY_MEAN = "monthly_mean"  # Verified monthly mean water level data for the station.
    ONE_MINUTE_WATER_LEVEL = (
        "one_minute_water_level"
        # One minute water level data for the station.
    )
    PREDICTIONS = "predictions"  # 6 minute predictions water level data for the station.*
    DATUMS = "datums"  # datums data for the stations.
    CURRENTS = "currents"  # Currents data for currents stations.
    CURRENTS_PREDICTIONS = (
        "currents_predictions"
        # Currents predictions data for currents predictions stations.
    )


class COOPS_Interval(StationDataInterval):  # noqa: N801
    H = "h"  # Hourly Met data and harmonic predictions will be returned
    HILO = "hilo"  # High/Low tide predictions for all stations.


class COOPS_TidalDatum(StationDatum):  # noqa: N801
    CRD = "CRD"  # Columbia River Datum
    IGLD = "IGLD"  # International Great Lakes Datum
    LWD = "LWD"  # Great Lakes Low Water Datum (Chart Datum)
    MHHW = "MHHW"  # Mean Higher High Water
    MHW = "MHW"  # Mean High Water
    MTL = "MTL"  # Mean Tide Level
    MSL = "MSL"  # Mean Sea Level
    MLW = "MLW"  # Mean Low Water
    MLLW = "MLLW"  # Mean Lower Low Water
    NAVD = "NAVD"  # North American Vertical Datum
    STND = "STND"  # Station Datum


class COOPS_VelocityType(Enum):  # noqa: N801
    SPEED_DIR = "speed_dir"  # Return results for speed and dirction
    DEFAULT = "default"  # Return results for velocity major, mean flood direction and mean ebb dirction


class COOPS_Units(Enum):  # noqa: N801
    METRIC = "metric"
    ENGLISH = "english"


class COOPS_TimeZone(Enum):  # noqa: N801
    GMT = "gmt"  # Greenwich Mean Time
    LST = "lst"  # Local Standard Time. The time local to the requested station.
    LST_LDT = "lst_ldt"  # Local Standard/Local Daylight Time. The time local to the requested station.


class COOPS_Station(Station):  # noqa: N801
    """
    a specific CO-OPS station
    """

    def __init__(self, id: Union[int, str]):
        """
        :param id: NOS ID, NWS ID, or station name

        from NOS ID:

        >>> COOPS_Station(1612480)
        COOPS_Station(1612480)

        from NWS ID:

        >>> COOPS_Station('OOUH1')
        COOPS_Station(1612340)

        from station name:

        >>> COOPS_Station('San Mateo Bridge')
        COOPS_Station(9414458)
        """

        stations = coops_stations()
        if id in stations.index:
            metadata = stations.loc[id]
        elif id in stations["nws_id"].values:
            metadata = stations[stations["nws_id"] == id]
        elif id in stations["name"].values:
            metadata = stations[stations["name"] == id]
        else:
            metadata = None

        if metadata is None or len(metadata) == 0:
            raise ValueError(f'station with "{id}" not found')

        removed = metadata["removed"]
        if isinstance(removed, Series):
            self.__active = pandas.isna(removed).any()
            removed = pandas.unique(removed.dropna().sort_values(ascending=False))
        else:
            self.__active = pandas.isna(removed)
        self.__removed = removed

        if isinstance(metadata, DataFrame):
            metadata = metadata.iloc[0]

        Station.__init__(self, id=str(metadata.name), location=metadata.geometry)

        self.nws_id = metadata["nws_id"]
        self.state = metadata["state"]
        self.name = metadata["name"]

        self.__query = None

    @property
    def current(self) -> bool:
        return self.__active

    @property
    def removed(self) -> Series:
        return self.__removed

    @property
    @lru_cache(maxsize=1)
    def constituents(self) -> DataFrame:
        """
        :return: table of tidal constituents for the current station
        """

        url = f"https://tidesandcurrents.noaa.gov/harcon.html?id={self.id}"
        response = requests.get(url)
        soup = BeautifulSoup(response.content, features="html.parser")
        table = soup.find_all("table", {"class": "table table-striped"})
        if len(table) > 0:
            table = table[0]
            columns = [field.text for field in table.find("thead").find("tr").find_all("th")]
            constituents = []
            for row in table.find_all("tr")[1:]:
                constituents.append([entry.text for entry in row.find_all("td")])
            constituents = DataFrame(constituents, columns=columns)
            constituents.rename(columns={"Constituent #": "#"}, inplace=True)
            constituents = constituents.astype(
                {
                    "#": numpy.int32,
                    "Amplitude": numpy.float64,
                    "Phase": numpy.float64,
                    "Speed": numpy.float64,
                }
            )
        else:
            constituents = DataFrame(columns=["#", "Amplitude", "Phase", "Speed"])

        constituents.set_index("#", inplace=True)
        return constituents

    def product(
        self,
        product: COOPS_Product,
        start_date: datetime,
        end_date: datetime = None,
        interval: COOPS_Interval = None,
        datum: COOPS_TidalDatum = None,
    ) -> Dataset:
        """
        retrieve data for the current station within the specified parameters

        :param product: CO-OPS product
        :param start_date: start date
        :param end_date: end date
        :param interval: time interval of data
        :param datum: tidal datum
        :return: data for the current station within the specified parameters

        >>> station = COOPS_Station(8632200)
        >>> station.product('water_level', start_date=datetime(2018, 9, 13), end_date=datetime(2018, 9, 16, 12))
        <xarray.Dataset>
        Dimensions:  (nos_id: 1, t: 841)
        Coordinates:
          * nos_id   (nos_id) int64 8632200
          * t        (t) datetime64[ns] 2018-09-13 ... 2018-09-16T12:00:00
            nws_id   (nos_id) <U5 'KPTV2'
            x        (nos_id) float64 -76.0
            y        (nos_id) float64 37.16
        Data variables:
            v        (nos_id, t) float32 1.67 1.694 1.73 1.751 ... 1.597 1.607 1.605
            s        (nos_id, t) float32 0.026 0.027 0.034 0.03 ... 0.018 0.019 0.021
            f        (nos_id, t) object '0,0,0,0' '0,0,0,0' ... '0,0,0,0' '0,0,0,0'
            q        (nos_id, t) object 'v' 'v' 'v' 'v' 'v' 'v' ... 'v' 'v' 'v' 'v' 'v'
        """

        logger.debug("Downloading: %s", self)
        if self.__query is None:
            self.__query = COOPS_Query(
                station=int(self.id),
                product=product,
                start_date=start_date,
                end_date=end_date,
                datum=datum,
                interval=interval,
            )
        else:
            if start_date is not None:
                self.__query.start_date = start_date
            if end_date is not None:
                self.__query.end_date = end_date
            if product is not None:
                self.__query.product = product
            if datum is not None:
                self.__query.datum = datum
            if interval is not None:
                self.__query.interval = interval

        data = self.__query.data

        data["nos_id"] = self.id
        data.set_index(["nos_id", data.index], inplace=True)

        data = data.to_xarray()

        if len(data["t"]) > 0:
            data = data.assign_coords(
                nws_id=("nos_id", [self.nws_id]),
                x=("nos_id", [self.location.x]),
                y=("nos_id", [self.location.y]),
            )
        else:
            data = data.assign_coords(
                nws_id=("nos_id", []),
                x=("nos_id", []),
                y=("nos_id", []),
            )

        return data

    def __str__(self) -> str:
        return f"{self.__class__.__name__} - {self.id} ({self.name}) - {self.location}"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id})"


class COOPS_Query(StationQuery):  # noqa: N801
    """
    abstraction of an individual query to the CO-OPS API
    https://api.tidesandcurrents.noaa.gov/api/prod/
    """

    URL = "https://tidesandcurrents.noaa.gov/api/datagetter?"

    def __init__(
        self,
        station: int,
        product: COOPS_Product,
        start_date: datetime,
        end_date: datetime = None,
        datum: COOPS_TidalDatum = None,
        units: COOPS_Units = None,
        time_zone: COOPS_TimeZone = None,
        interval: COOPS_Interval = None,
    ):
        """
        instantiate a new query with the specified parameters

        :param station: NOS station ID
        :param product: CO-OPS product
        :param start_date: start date
        :param end_date: end date
        :param datum: station datum
        :param units: one of `metric`, `english`
        :param time_zone: time zone of data
        :param interval: time interval of data
        :return: data for the current station within the specified parameters

        >>> COOPS_Query(1612480, start_date='2022-01-01', end_date='2022-01-03')
        COOPS_Query(1612480, datetime.datetime(2022, 1, 1, 0, 0), datetime.datetime(2022, 1, 3, 0, 0), 'water_level', 'MLLW', 'metric', 'gmt', 'h')
        """

        if isinstance(station, COOPS_Station):
            station = station.id
        if end_date is None:
            end_date = datetime.today()
        if datum is None:
            datum = COOPS_TidalDatum.STND
        if units is None:
            units = COOPS_Units.METRIC
        if time_zone is None:
            time_zone = COOPS_TimeZone.GMT
        if interval is None:
            interval = COOPS_Interval.H

        StationQuery.__init__(
            self, station_id=station, product=product, start_date=start_date, end_date=end_date  # type: ignore[arg-type]
        )

        self.station_id = station
        self.product = product
        self.start_date = start_date
        self.end_date = end_date
        self.datum = datum
        self.units = units
        self.time_zone = time_zone
        self.interval = interval

        self.__previous_query = None
        self.__error = None
        self.__data = None

    @property
    def start_date(self) -> datetime:
        return self.__start_date

    @start_date.setter
    def start_date(self, start_date: datetime) -> None:
        self.__start_date = pandas.to_datetime(start_date)

    @property
    def end_date(self) -> datetime:
        return self.__end_date

    @end_date.setter
    def end_date(self, end_date: datetime) -> None:
        self.__end_date = pandas.to_datetime(end_date)

    @property
    def product(self) -> COOPS_Product:
        return self.__product

    @product.setter
    def product(self, product: COOPS_Product) -> None:
        self.__product = typepigeon.convert_value(product, COOPS_Product)

    @property
    def datum(self) -> COOPS_TidalDatum:
        return self.__datum

    @datum.setter
    def datum(self, datum: COOPS_TidalDatum) -> None:
        self.__datum = typepigeon.convert_value(datum, COOPS_TidalDatum)

    @property
    def units(self) -> COOPS_Units:
        return self.__units

    @units.setter
    def units(self, units: COOPS_Units) -> None:
        self.__units = typepigeon.convert_value(units, COOPS_Units)

    @property
    def time_zone(self) -> COOPS_TimeZone:
        return self.__time_zone

    @time_zone.setter
    def time_zone(self, time_zone: COOPS_TimeZone) -> None:
        self.__time_zone = typepigeon.convert_value(time_zone, COOPS_TimeZone)

    @property
    def interval(self) -> COOPS_Interval:
        return self.__interval

    @interval.setter
    def interval(self, interval: COOPS_Interval) -> None:
        self.__interval = typepigeon.convert_value(interval, COOPS_Interval)

    @property
    def query(self) -> Dict[str, Any]:
        self.__error = None

        product = self.product
        if isinstance(product, Enum):
            product = product.value
        start_date = self.start_date
        if start_date is not None and not isinstance(start_date, str):
            start_date = f"{self.start_date:%Y%m%d %H:%M}"
        datum = self.datum
        if isinstance(datum, Enum):
            datum = datum.value
        units = self.units
        if isinstance(units, Enum):
            units = units.value
        time_zone = self.time_zone
        if isinstance(time_zone, Enum):
            time_zone = time_zone.value
        interval = self.interval
        if isinstance(interval, Enum):
            interval = interval.value

        return {
            "station": self.station_id,
            "product": product,
            "begin_date": start_date,
            "end_date": f"{self.end_date:%Y%m%d %H:%M}",
            "datum": datum,
            "units": units,
            "time_zone": time_zone,
            "interval": interval,
            "format": "json",
            "application": "noaa/nos/csdl/stormevents",
        }

    @property
    def data(self) -> DataFrame:
        """
        :return: data for the current query parameters

        >>> query = COOPS_Query(1612480, 'water_level', start_date='2022-01-01', end_date='2022-01-03')
        >>> query.data
                                 v      s        f  q
        t
        2022-01-01 00:00:00  1.193  0.002  0,0,0,0  v
        2022-01-01 00:06:00  1.180  0.002  0,0,0,0  v
        2022-01-01 00:12:00  1.167  0.002  0,0,0,0  v
        2022-01-01 00:18:00  1.156  0.003  0,0,0,0  v
        2022-01-01 00:24:00  1.147  0.003  0,0,0,0  v
                            ...    ...      ... ..
        2022-01-02 23:36:00  1.229  0.004  0,0,0,0  v
        2022-01-02 23:42:00  1.219  0.003  0,0,0,0  v
        2022-01-02 23:48:00  1.223  0.004  0,0,0,0  v
        2022-01-02 23:54:00  1.217  0.004  0,0,0,0  v
        2022-01-03 00:00:00  1.207  0.002  0,0,0,0  v
        [481 rows x 4 columns]
        """

        if self.__previous_query is None or self.query != self.__previous_query:
            response = requests.get(self.URL, params=self.query)
            data = response.json()
            fields = ["t", "v", "s", "f", "q"]
            if "error" in data or "data" not in data:
                if "error" in data:
                    self.__error = data["error"]["message"]
                data = DataFrame(columns=fields)
            else:
                data = DataFrame(data["data"], columns=fields)
                data[data == ""] = numpy.nan
                data = data.astype(
                    {"v": numpy.float32, "s": numpy.float32, "f": "string", "q": "string"},
                    errors="ignore",
                )
                data["t"] = pandas.to_datetime(data["t"])

            data.set_index("t", inplace=True)
            self.__data = data
        return self.__data

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({", ".join(repr(value) for value in (self.station_id, self.start_date, self.end_date, self.product.value, self.datum.value, self.units.value, self.time_zone.value, self.interval.value))})'


@lru_cache(maxsize=1)
def __coops_stations_html_tables() -> element.ResultSet:
    url = "https://access.co-ops.nos.noaa.gov/nwsproducts.html?type=current"
    logger.debug("Downloading: %s", url)
    response = requests.get(url)
    soup = BeautifulSoup(response.content, features="html.parser")
    return soup.find_all("div", {"class": "table-responsive"})


@lru_cache(maxsize=1)
def coops_stations(station_status: StationStatus = None) -> GeoDataFrame:
    """
    retrieve a list of CO-OPS stations with associated metadata

    :param station_status: one of ``active`` or ``discontinued``
    :return: data frame of stations

    >>> coops_stations()
            nws_id                              name state        status                                            removed                     geometry
    nos_id
    1600012  46125                         QREB buoy              active                                               <NA>   POINT (122.62500 37.75000)
    8735180  DILA1                    Dauphin Island    AL        active  2019-07-18 10:00:00,2018-07-30 16:40:00,2017-0...   POINT (-88.06250 30.25000)
    8557380  LWSD1                             Lewes    DE        active  2019-08-01 00:00:00,2018-06-18 00:00:00,2017-0...   POINT (-75.12500 38.78125)
    8465705  NWHC3                         New Haven    CT        active  2019-08-18 14:55:00,2019-08-18 14:54:00,2018-0...   POINT (-72.93750 41.28125)
    9439099  WAUO3                             Wauna    OR        active  2019-08-19 22:59:00,2014-06-20 21:30:00,2013-0...  POINT (-123.43750 46.15625)
    ...        ...                               ...   ...           ...                                                ...                          ...
    8448725  MSHM3               Menemsha Harbor, MA    MA  discontinued  2013-09-26 23:59:00,2013-09-26 00:00:00,2012-0...   POINT (-70.75000 41.34375)
    8538886  TPBN4             Tacony-Palmyra Bridge    NJ  discontinued  2013-11-11 00:01:00,2013-11-11 00:00:00,2012-0...   POINT (-75.06250 40.00000)
    9439011  HMDO3                           Hammond    OR  discontinued  2014-08-13 00:00:00,2011-04-12 23:59:00,2011-0...  POINT (-123.93750 46.18750)
    8762372  LABL1  East Bank 1, Norco, B. LaBranche    LA  discontinued  2012-11-05 10:38:00,2012-11-05 10:37:00,2012-1...   POINT (-90.37500 30.04688)
    8530528  CARN4       CARLSTADT, HACKENSACK RIVER    NJ  discontinued            1994-11-12 23:59:00,1994-11-12 00:00:00   POINT (-74.06250 40.81250)
    [436 rows x 6 columns]
    >>> coops_stations(station_status='active')
            nws_id                          name state  status removed                     geometry
    nos_id
    1600012  46125                     QREB buoy        active    <NA>   POINT (122.62500 37.75000)
    1611400  NWWH1                    Nawiliwili    HI  active    <NA>  POINT (-159.37500 21.95312)
    1612340  OOUH1                      Honolulu    HI  active    <NA>  POINT (-157.87500 21.31250)
    1612480  MOKH1                      Mokuoloe    HI  active    <NA>  POINT (-157.75000 21.43750)
    1615680  KLIH1       Kahului, Kahului Harbor    HI  active    <NA>  POINT (-156.50000 20.89062)
    ...        ...                           ...   ...     ...     ...                          ...
    9759394  MGZP4                      Mayaguez    PR  active    <NA>   POINT (-67.18750 18.21875)
    9759938  MISP4                   Mona Island        active    <NA>   POINT (-67.93750 18.09375)
    9761115  BARA9                       Barbuda        active    <NA>   POINT (-61.81250 17.59375)
    9999530  FRCB6  Bermuda, Ferry Reach Channel        active    <NA>   POINT (-64.68750 32.37500)
    9999531               Calcasieu Test Station    LA  active    <NA>   POINT (-93.31250 29.76562)
    [365 rows x 6 columns]
    >>> coops_stations(station_status='discontinued')
            nws_id                               name state        status                                            removed                     geometry
    nos_id
    8530528  CARN4        CARLSTADT, HACKENSACK RIVER    NJ  discontinued            1994-11-12 23:59:00,1994-11-12 00:00:00   POINT (-74.06250 40.81250)
    9415064  NCHC1         ANTIOCH, SAN JOAQUIN RIVER    CA  discontinued            1997-03-03 23:59:00,1997-03-03 00:00:00  POINT (-121.81250 38.03125)
    9415316  RVXC1                          Rio Vista    CA  discontinued            1997-03-04 23:59:00,1997-03-04 00:00:00  POINT (-121.68750 38.15625)
    9440572  ILWW1            JETTY A, COLUMBIA RIVER    WA  discontinued                                1997-04-11 23:00:00  POINT (-124.06250 46.28125)
    8760551  SPSL1                         South Pass    LA  discontinued  2000-09-26 23:59:00,2000-09-26 00:00:00,1998-1...   POINT (-89.12500 28.98438)
    ...        ...                                ...   ...           ...                                                ...                          ...
    8726667  MCYF1                 Mckay Bay Entrance    FL  discontinued  2020-05-20 00:00:00,2019-03-08 00:00:00,2017-0...   POINT (-82.43750 27.90625)
    8772447  FCGT2                           Freeport    TX  discontinued  2020-05-24 18:45:00,2018-10-10 21:50:00,2018-1...   POINT (-95.31250 28.93750)
    9087079  GBWW3                          Green Bay    WI  discontinued  2020-10-28 13:00:00,2007-08-06 23:59:00,2007-0...   POINT (-88.00000 44.53125)
    8770570  SBPT2                  Sabine Pass North    TX  discontinued  2021-01-18 00:00:00,2020-09-30 15:45:00,2020-0...   POINT (-93.87500 29.73438)
    8740166  GBRM6  Grand Bay NERR, Mississippi Sound    MS  discontinued  2022-04-07 00:00:00,2022-03-30 23:58:00,2015-1...   POINT (-88.37500 30.40625)
    [71 rows x 6 columns]
    """

    tables = __coops_stations_html_tables()

    status_tables = {
        StationStatus.ACTIVE: (0, "NWSTable"),
        StationStatus.DISCONTINUED: (1, "HistNWSTable"),
    }

    dataframes = {}
    for status, (table_index, table_id) in status_tables.items():
        table = tables[table_index]
        table = table.find("table", {"id": table_id}).find_all("tr")
        stations_columns = [field.text for field in table[0].find_all("th")]
        stations = DataFrame(
            [[value.text.strip() for value in station.find_all("td")] for station in table[1:]],
            columns=stations_columns,
        )
        stations.rename(
            columns={
                "NOS ID": "nos_id",
                "NWS ID": "nws_id",
                "Latitude": "y",
                "Longitude": "x",
                "State": "state",
                "Station Name": "name",
            },
            inplace=True,
        )
        stations = stations.astype(
            {
                "nos_id": numpy.int32,
                "nws_id": "string",
                "x": numpy.float32,
                "y": numpy.float32,
                "state": "string",
                "name": "string",
            },
            copy=False,
        )
        stations.set_index("nos_id", inplace=True)

        if status == StationStatus.DISCONTINUED:
            with (Path(__file__).parent / "us_states.json").open() as us_states_file:
                us_states = json.load(us_states_file)

            for name, abbreviation in us_states.items():
                stations.loc[stations["state"] == name, "state"] = abbreviation

            stations.rename(columns={"Removed Date/Time": "removed"}, inplace=True)

            stations["removed"] = pandas.to_datetime(stations["removed"]).astype("string")

            stations = (
                stations[~pandas.isna(stations["removed"])]
                .sort_values("removed", ascending=False)
                .groupby("nos_id")
                .aggregate(
                    {
                        "nws_id": "first",
                        "removed": ",".join,
                        "y": "first",
                        "x": "first",
                        "state": "first",
                        "name": "first",
                    }
                )
            )
        else:
            stations["removed"] = pandas.NA

        stations["status"] = status.value
        dataframes[status] = stations

    active_stations = dataframes[StationStatus.ACTIVE]
    discontinued_stations = dataframes[StationStatus.DISCONTINUED]
    discontinued_stations.loc[
        discontinued_stations.index.isin(active_stations.index), "status"
    ] = StationStatus.ACTIVE.value

    stations = pandas.concat(
        (
            active_stations[~active_stations.index.isin(discontinued_stations.index)],
            discontinued_stations,
        )
    )
    stations.loc[pandas.isna(stations["status"]), "status"] = StationStatus.ACTIVE.value
    stations.sort_values(["status", "removed"], na_position="first", inplace=True)

    if station_status is not None:
        if isinstance(station_status, StationStatus):
            station_status = station_status.value
        stations = stations[stations["status"] == station_status]

    return GeoDataFrame(
        stations[["nws_id", "name", "state", "status", "removed"]],
        geometry=geopandas.points_from_xy(stations["x"], stations["y"]),
    )


def coops_stations_within_region(
    region: Polygon,
    station_status: StationStatus = None,
) -> GeoDataFrame:
    """
    retrieve all stations within the specified region of interest

    :param region: polygon or multipolygon denoting region of interest
    :param station_status: one of ``active`` or ``discontinued``
    :return: data frame of stations within the specified region

    >>> import shapely
    >>> east_coast = shapely.geometry.box(-85, 25, -65, 45)
    >>> coops_stations_within_region(east_coast)
            nws_id                           name state        status                                            removed                    geometry
    nos_id
    8726412  MTBF1               Middle Tampa Bay              active                                               <NA>  POINT (-82.62500 27.65625)
    8726679  TSHF1              East Bay Causeway    FL        active                                               <NA>  POINT (-82.43750 27.92188)
    8726694  TPAF1          TPA Cruise Terminal 2    FL        active                                               <NA>  POINT (-82.43750 27.93750)
    9044036  FWNM4                     Fort Wayne    MI        active  2005-04-29 23:59:00,2005-04-29 00:00:00,2001-1...  POINT (-83.06250 42.31250)
    9075035  ESVM4                     Essexville    MI        active  2007-03-28 23:59:00,2007-03-28 00:00:00,2007-0...  POINT (-83.87500 43.62500)
    ...        ...                            ...   ...           ...                                                ...                         ...
    8720503  GCVF1  Red Bay Point, St Johns River    FL  discontinued  2017-10-07 20:54:00,2017-10-07 10:54:00,2017-1...  POINT (-81.62500 29.98438)
    8654400  CFPN7     Cape Hatteras Fishing Pier    NC  discontinued  2018-09-19 23:59:00,2003-09-18 23:59:00,2003-0...  POINT (-75.62500 35.21875)
    8720625  RCYF1     Racy Point, St Johns River    FL  discontinued  2019-08-05 14:00:00,2017-06-14 15:36:00,2017-0...  POINT (-81.56250 29.79688)
    8423898  FTPN3                     Fort Point    NH  discontinued  2020-04-13 00:00:00,2014-08-05 00:00:00,2012-0...  POINT (-70.68750 43.06250)
    8726667  MCYF1             Mckay Bay Entrance    FL  discontinued  2020-05-20 00:00:00,2019-03-08 00:00:00,2017-0...  POINT (-82.43750 27.90625)
    [164 rows x 6 columns]
    """

    stations = coops_stations(station_status=station_status)
    if region:
        return stations[stations.within(region)]
    else:
        return stations


def coops_stations_within_bounds(
    minx: float,
    miny: float,
    maxx: float,
    maxy: float,
    station_status: StationStatus = None,
) -> GeoDataFrame:
    return coops_stations_within_region(
        region=shapely.geometry.box(minx=minx, miny=miny, maxx=maxx, maxy=maxy),
        station_status=station_status,
    )


def coops_product_within_region(
    product: COOPS_Product,
    region: Union[Polygon, MultiPolygon],
    start_date: datetime,
    end_date: datetime = None,
    datum: COOPS_TidalDatum = None,
    interval: COOPS_Interval = None,
    station_status: StationStatus = None,
) -> Dataset:
    """
    retrieve CO-OPS data from within the specified region of interest

    :param product: CO-OPS product; one of ``water_level``, ``air_temperature``, ``water_temperature``, ``wind``, ``air_pressure``, ``air_gap``, ``conductivity``, ``visibility``, ``humidity``, ``salinity``, ``hourly_height``, ``high_low``, ``daily_mean``, ``monthly_mean``, ``one_minute_water_level``, ``predictions``, ``datums``, ``currents``, ``currents_predictions``
    :param region: polygon or multipolygon denoting region of interest
    :param start_date: start date of CO-OPS query
    :param end_date: start date of CO-OPS query
    :param datum: tidal datum
    :param interval: data time interval
    :param station_status: either ``active`` or ``discontinued``
    :return: array of data within the specified region

    >>> from datetime import datetime, timedelta
    >>> import shapely
    >>> east_coast = shapely.geometry.box(-85, 25, -65, 45)
    >>> coops_product_within_region('water_level', region=east_coast, start_date=datetime(2022, 4, 2, 12), end_date=datetime(2022, 4, 2, 12, 30))
    <xarray.Dataset>
    Dimensions:  (t: 6, nos_id: 111)
    Coordinates:
      * t        (t) datetime64[ns] 2022-04-02T12:00:00 ... 2022-04-02T12:30:00
      * nos_id   (nos_id) int64 9044036 9075035 9052076 ... 8516945 9052000 8638901
        nws_id   (nos_id) <U5 'FWNM4' 'ESVM4' 'OCTN6' ... 'KPTN6' 'CAVN6' 'CHBV2'
        x        (nos_id) float64 -83.06 -83.88 -78.75 ... -73.75 -76.31 -76.06
        y        (nos_id) float64 42.31 43.62 43.34 42.09 ... 40.81 44.12 37.03
    Data variables:
        v        (nos_id, t) float32 175.1 175.1 175.1 175.1 ... 2.084 2.089 2.096
        s        (nos_id, t) float32 0.0 0.0 0.0 0.0 0.0 ... 0.059 0.076 0.088 0.08
        f        (nos_id, t) object '0,0,0,0' '0,0,0,0' ... '1,0,0,0' '1,0,0,0'
        q        (nos_id, t) object 'p' 'p' 'p' 'p' 'p' 'p' ... 'p' 'p' 'p' 'p' 'p'
    """

    stations = coops_stations_within_region(region=region, station_status=station_status)
    station_data = [
        COOPS_Station(station).product(
            product=product,
            start_date=start_date,
            end_date=end_date,
            datum=datum,
            interval=interval,
        )
        for station in stations.index
    ]
    station_data = [station for station in station_data if len(station["t"]) > 0]
    return xarray.combine_nested(station_data, concat_dim="nos_id")


get_coops_stations = coops_stations_within_region



import itertools
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Tuple
from typing import TypeVar
from typing import Union

import xarray as xr
from shapely.geometry import box
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon

from . import models
from .custom_types import ScalarOrArray

_T = TypeVar("_T")
_U = TypeVar("_U")


# https://gis.stackexchange.com/questions/201789/verifying-formula-that-will-convert-longitude-0-360-to-180-to-180
# lon1 is the longitude varying from -180 to 180 or 180W-180E
# lon3 is the longitude variable from 0 to 360 (all positive)


def lon1_to_lon3(lon1: ScalarOrArray) -> ScalarOrArray:
    return lon1 % 360


def lon3_to_lon1(lon3: ScalarOrArray) -> ScalarOrArray:
    return ((lon3 + 180) % 360) - 180


def get_region_from_bbox_corners(
    lon_min: Union[float, None] = None,
    lon_max: Union[float, None] = None,
    lat_min: Union[float, None] = None,
    lat_max: Union[float, None] = None,
    symmetric: bool = True,
) -> Polygon:
    """Return a ``shapely.geometry.Polygon`` from the corners of a Bounding Box."""
    bbox_kwargs = {
        "lon_min": lon_min,
        "lon_max": lon_max,
        "lat_min": lat_min,
        "lat_max": lat_max,
    }
    if symmetric:
        klass = models.SymmetricBBox
    else:
        klass = models.AsymmetricBBox

    # Create the BBox (i.e. validate input values)
    # Not the most beautiful code in the world, but Pydantic needs keyword arguments
    # and we want to only pass arguments that have actual values and not None
    # moreover, we create a new object because mypy was complaining about redefining
    filtered_bbox_kwargs: dict[str, float] = {k: v for (k, v) in bbox_kwargs.items() if v}
    bbox = klass(**filtered_bbox_kwargs)

    # Create the region
    region = box(
        minx=bbox.lon_min,
        miny=bbox.lat_min,
        maxx=bbox.lon_max,
        maxy=bbox.lat_max,
    )
    return region


def get_region(
    region: Union[Union[Polygon, MultiPolygon], None] = None,
    lon_min: Union[float, None] = None,
    lon_max: Union[float, None] = None,
    lat_min: Union[float, None] = None,
    lat_max: Union[float, None] = None,
    symmetric: bool = True,
) -> Union[Polygon, None]:
    """
    Return a shapely Region

    The region can either be a ``shapely.geometry`` object or it can be defined
    from the corners of a Bounding Box.

    If both a ``region`` and the corners are specified then an Exception is raised.

    - If ``symmetric`` is ``False``, then it is assumed that Longitude ∈ [0, 360).
    - If ``symmetric`` is ``True``, then it is assumed that Longitude ∈ [-180, 180).
    """
    if any((lon_min, lon_max, lat_min, lat_max)):
        # Ensure that only one of region and bbox have been specified, not both
        if region:
            msg = "You must specify either `region` or the `BBox` corners, not both"
            raise ValueError(msg)

        region = get_region_from_bbox_corners(
            lon_min=lon_min,
            lon_max=lon_max,
            lat_min=lat_min,
            lat_max=lat_max,
            symmetric=symmetric,
        )
    return region


# https://docs.python.org/3/library/itertools.html#itertools-recipes
# https://github.com/more-itertools/more-itertools/blob/2ff5943d76afa4591b5b4ae8cb4524578d365f67/more_itertools/recipes.pyi#L42-L48
def grouper(
    iterable: Iterable[_T],
    n: int,
    *,
    incomplete: str = "fill",
    fillvalue: Union[_U, None] = None,
) -> Iterator[Tuple[Union[_T, _U], ...]]:
    """Collect data into non-overlapping fixed-length chunks or blocks"""
    # grouper('ABCDEFG', 3, fillvalue='x') --> ABC DEF Gxx
    # grouper('ABCDEFG', 3, incomplete='strict') --> ABC DEF ValueError
    # grouper('ABCDEFG', 3, incomplete='ignore') --> ABC DEF
    args = [iter(iterable)] * n
    if incomplete == "fill":
        return itertools.zip_longest(*args, fillvalue=fillvalue)
    if incomplete == "strict":
        return zip(*args, strict=True)  # type: ignore[call-overload]
    if incomplete == "ignore":
        return zip(*args)
    else:
        raise ValueError("Expected fill, strict, or ignore")


def merge_datasets(datasets: List[xr.Dataset], size: int = 5) -> List[xr.Dataset]:
    datasets = [xr.merge(g for g in group if g) for group in grouper(datasets, size)]
    return datasets



from __future__ import annotations

import datetime
from enum import Enum

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon

from . import coops
from . import ioc


STATIONS_COLUMNS: list[str] = [
    "provider",
    "provider_id",
    "country",
    "location",
    "lon",
    "lat",
    "is_active",
    "start_date",
    "last_observation",
    "geometry",
]


class Provider(str, Enum):
    ALL: str = "ALL"
    COOPS: str = "COOPS"
    IOC: str = "IOC"


def _get_ioc_stations(
    activity_threshold: datetime.timedelta,
    region: Polygon | MultiPolygon | None = None,
) -> gpd.GeoDataFrame:
    # Convert activity threshold to a Timezone Aware Datetime object
    now_utc = datetime.datetime.now(datetime.timezone.utc)
    activity_threshold_ts = now_utc - activity_threshold

    # Get full metadata
    ioc_gdf = ioc.get_ioc_stations(region=region)

    # Normalize IOC
    # Convert delay to minutes
    ioc_gdf = ioc_gdf.assign(
        delay=pd.concat(
            (
                ioc_gdf.delay[ioc_gdf.delay.str.endswith("'")].str[:-1].astype(int),
                ioc_gdf.delay[ioc_gdf.delay.str.endswith("h")].str[:-1].astype(int) * 60,
                ioc_gdf.delay[ioc_gdf.delay.str.endswith("d")].str[:-1].astype(int) * 24 * 60,
            )
        )
    )

    # Some IOC stations appear to have negative delay due to server
    # [time drift](https://www.bluematador.com/docs/troubleshooting/time-drift-ntp)
    # or for other Provider-specific reasons. IOC suggests to ignore the negative
    # delay and consider the stations as active.
    # https://github.com/oceanmodeling/searvey/issues/40#issuecomment-1219509512
    ioc_gdf.loc[(ioc_gdf.delay < 0), "delay"] = 0

    # Calculate the timestamp of the last observation
    ioc_gdf = ioc_gdf.assign(
        last_observation=now_utc - ioc_gdf.delay.apply(lambda x: datetime.timedelta(minutes=x))
    )

    ioc_gdf = ioc_gdf.assign(
        provider=Provider.IOC.value,
        provider_id=ioc_gdf.ioc_code,
        start_date=ioc_gdf.added_to_system.apply(pd.to_datetime).dt.tz_localize("UTC"),
        is_active=ioc_gdf.last_observation > activity_threshold_ts,
    )

    # Filter out columns
    ioc_gdf = ioc_gdf[STATIONS_COLUMNS]

    return ioc_gdf


def _get_coops_stations(
    region: Polygon | MultiPolygon | None = None,
) -> gpd.GeoDataFrame:
    coops_gdf = coops.coops_stations_within_region(region=region)
    coops_gdf = coops_gdf.assign(
        provider=Provider.COOPS.value,
        provider_id=coops_gdf.index,
        country=np.where(coops_gdf.state.str.len() > 0, "USA", None),  # type: ignore[call-overload]
        location=coops_gdf[["name", "state"]].agg(", ".join, axis=1),
        lon=coops_gdf.geometry.x,
        lat=coops_gdf.geometry.y,
        is_active=coops_gdf.status == "active",
        start_date=pd.NaT,
        last_observation=coops_gdf[coops_gdf.status == "discontinued"]
        .removed.str[:18]
        .apply(pd.to_datetime)
        .dt.tz_localize("UTC"),
    )[STATIONS_COLUMNS]
    return coops_gdf


def get_stations(
    activity_threshold: datetime.timedelta = datetime.timedelta(days=3),
    providers: list[Provider] = Provider.ALL,
    region: Polygon | MultiPolygon | None = None,
) -> gpd.GeoDataFrame:
    dataframes = []
    if Provider.ALL in providers or Provider.IOC in providers:
        dataframes.append(_get_ioc_stations(activity_threshold=activity_threshold, region=region))
    if Provider.ALL in providers or Provider.COOPS in providers:
        dataframes.append(_get_coops_stations(region=region))
    df = pd.concat(dataframes).reset_index(drop=True)
    return df



"""
For the IOC stations we parse their website_.

:: _website: http://www.ioc-sealevelmonitoring.org/list.php?showall=all

This page contains 3 tables:

- ``general``
- ``contacts``
- ``performance``

We parse all 3 of them and we merge them.
"""
from __future__ import annotations

import datetime
import functools
import logging
import warnings
from typing import Optional
from typing import Union

import bs4
import geopandas as gpd
import html5lib  # noqa: F401  # imported but unused
import limits
import lxml  # noqa: F401  # imported but unused
import pandas as pd
import requests
import xarray as xr
from shapely.geometry import MultiPolygon
from shapely.geometry import Polygon

from .multi import multiprocess
from .multi import multithread
from .rate_limit import RateLimit
from .rate_limit import wait
from .utils import get_region
from .utils import merge_datasets


logger = logging.getLogger(__name__)

# constants
IOC_RATE_LIMIT = limits.parse("5/second")
IOC_MAX_DAYS_PER_REQUEST = 30
IOC_BASE_URL = "http://www.ioc-sealevelmonitoring.org/bgraph.php?code={ioc_code}&output=tab&period={period}&endtime={endtime}"
IOC_STATIONS_KWARGS = [
    {"output": "general", "skip_table_rows": 4},
    {"output": "contacts", "skip_table_rows": 4},
    {"output": "performance", "skip_table_rows": 8},
]
IOC_STATIONS_COLUMN_NAMES = {
    "general": [
        "ioc_code",
        "gloss_id",
        "country",
        "location",
        "connection",
        "dcp_id",
        "last_observation_level",
        "last_observation_time",
        "delay",
        "interval",
        "view",
    ],
    "contacts": [
        "ioc_code",
        "gloss_id",
        "lat",
        "lon",
        "country",
        "location",
        "connection",
        "contacts",
        "view",
    ],
    "performance": [
        "ioc_code",
        "gloss_id",
        "country",
        "location",
        "connection",
        "added_to_system",
        "observations_arrived_per_week",
        "observations_expected_per_week",
        "observations_ratio_per_week",
        "observations_arrived_per_month",
        "observations_expected_per_month",
        "observations_ratio_per_month",
        "observations_ratio_per_day",
        "sample_interval",
        "average_delay_per_day",
        "transmit_interval",
        "view",
    ],
}
IOC_STATION_DATA_COLUMNS_TO_DROP = [
    "bat",
    "sw1",
    "sw2",
    "(m)",
]
IOC_STATION_DATA_COLUMNS = {
    "Time (UTC)": "time",
    "aqu(m)": "aqu",
    "atm(m)": "atm",
    "bat(V)": "bat",
    "bub(m)": "bub",
    "bwl(m)": "bwl",
    "ecs(m)": "ecs",
    "enb(m)": "enb",
    "enc(m)": "enc",
    "flt(m)": "flt",
    "pr1(m)": "pr1",
    "pr2(m)": "pr2",
    "prs(m)": "prs",
    "pwl(m)": "pwl",
    "ra2(m)": "ra2",
    "rad(m)": "rad",
    "stp(m)": "stp",
    "sw1(min)": "sw1",
    "sw2(min)": "sw2",
    "wls(m)": "wls",
}


def get_ioc_stations_by_output(output: str, skip_table_rows: int) -> pd.DataFrame:
    url = f"https://www.ioc-sealevelmonitoring.org/list.php?showall=all&output={output}#"
    logger.debug("Downloading: %s", url)
    response = requests.get(url)
    assert response.ok, f"failed to download: {url}"
    logger.debug("Downloaded: %s", url)
    soup = bs4.BeautifulSoup(response.content, "html5lib")
    logger.debug("Created soup: %s", url)
    table = soup.find("table", {"class": "nice"})
    trs = table.find_all("tr")
    table_contents = "\n".join(str(tr) for tr in trs[skip_table_rows:])
    html = f"<table>{table_contents}</table>"
    logger.debug("Created table: %s", url)
    df = pd.read_html(html)[0]
    logger.debug("Parsed table: %s", url)
    df.columns = IOC_STATIONS_COLUMN_NAMES[output]
    df = df.drop(columns="view")
    return df


def normalize_ioc_stations(df: pd.DataFrame) -> gpd.GeoDataFrame:
    df = df.assign(
        # fmt: off
        gloss_id=df.gloss_id.astype(pd.Int64Dtype()),
        observations_ratio_per_day=df.observations_ratio_per_day.replace("-", "0%").str[:-1].astype(int),
        observations_ratio_per_week=df.observations_ratio_per_week.replace("-", "0%").str[:-1].astype(int),
        observations_ratio_per_month=df.observations_ratio_per_month.replace("-", "0%").str[:-1].astype(int),
        # fmt: on
    )
    gdf = gpd.GeoDataFrame(
        data=df,
        geometry=gpd.points_from_xy(df.lon, df.lat, crs="EPSG:4326"),
    )
    return gdf


@functools.lru_cache
def _get_ioc_stations() -> gpd.GeoDataFrame:
    """
    Return IOC station metadata from: http://www.ioc-sealevelmonitoring.org/list.php?showall=all

    :return: ``pandas.DataFrame`` with the station metadata
    """

    # The IOC web page with the station metadata contains really ugly HTTP code.
    # creating the `bs4.Beautiful` instance takes as much time as you need to download the page,
    # therefore multiprocessing is actually faster than multithreading
    ioc_stations_results = multiprocess(
        func=get_ioc_stations_by_output,
        func_kwargs=IOC_STATIONS_KWARGS,
    )
    ioc_stations = functools.reduce(pd.merge, (r.result for r in ioc_stations_results))
    ioc_stations = normalize_ioc_stations(ioc_stations)
    return ioc_stations


def get_ioc_stations(
    region: Optional[Union[Polygon, MultiPolygon]] = None,
    lon_min: Optional[float] = None,
    lon_max: Optional[float] = None,
    lat_min: Optional[float] = None,
    lat_max: Optional[float] = None,
) -> gpd.GeoDataFrame:
    """
    Return IOC station metadata from: http://www.ioc-sealevelmonitoring.org/list.php?showall=all

    If `region` is defined then the stations that are outside of the region are
    filtered out.. If the coordinates of the Bounding Box are defined then
    stations outside of the BBox are filtered out. If both ``region`` and the
    Bounding Box are defined, then an exception is raised.

    Note: The longitudes of the IOC stations are in the [-180, 180] range.

    :param region: ``Polygon`` or ``MultiPolygon`` denoting region of interest
    :param lon_min: The minimum Longitude of the Bounding Box.
    :param lon_max: The maximum Longitude of the Bounding Box.
    :param lat_min: The minimum Latitude of the Bounding Box.
    :param lat_max: The maximum Latitude of the Bounding Box.
    :return: ``pandas.DataFrame`` with the station metadata
    """
    region = get_region(
        region=region,
        lon_min=lon_min,
        lon_max=lon_max,
        lat_min=lat_min,
        lat_max=lat_max,
        symmetric=True,
    )

    ioc_stations = _get_ioc_stations()
    if region:
        ioc_stations = ioc_stations[ioc_stations.within(region)]
    return ioc_stations


def normalize_ioc_station_data(ioc_code: str, df: pd.DataFrame, truncate_seconds: bool) -> pd.DataFrame:
    # Each station may have more than one sensors.
    # Some of the sensors have nothing to do with sea level height. We drop these sensors
    df = df.rename(columns=IOC_STATION_DATA_COLUMNS)
    logger.debug("%s: df contains the following columns: %s", ioc_code, df.columns)
    df = df.drop(columns=IOC_STATION_DATA_COLUMNS_TO_DROP, errors="ignore")
    if len(df.columns) == 1:
        # all the data columns have been dropped!
        msg = f"{ioc_code}: The table does not contain any sensor data!"
        logger.info(msg)
        raise ValueError(msg)
    df = df.assign(
        ioc_code=ioc_code,
        time=pd.to_datetime(df.time),
    )
    if truncate_seconds:
        # Truncate seconds from timestamps: https://stackoverflow.com/a/28783971/592289
        # WARNING: This can potentially lead to duplicates!
        df = df.assign(time=df.time.dt.floor("min"))
        if df.time.duplicated().any():
            # There are duplicates. Keep the first datapoint per minute.
            msg = f"{ioc_code}: Duplicate timestamps have been detected after the truncation of seconds. Keeping the first datapoint per minute"
            warnings.warn(msg)
            df = df.iloc[df.time.drop_duplicates().index].reset_index(drop=True)
    return df


def get_ioc_station_data(
    ioc_code: str,
    endtime: Union[str, datetime.date] = datetime.date.today(),
    period: float = IOC_MAX_DAYS_PER_REQUEST,
    truncate_seconds: bool = True,
    rate_limit: Optional[RateLimit] = None,
) -> pd.DataFrame:
    """Retrieve the TimeSeries of a single IOC station."""

    if rate_limit:
        while rate_limit.reached(identifier="IOC"):
            wait()

    endtime = pd.to_datetime(endtime).date().isoformat()
    url = IOC_BASE_URL.format(ioc_code=ioc_code, endtime=endtime, period=period)
    logger.info("%s: Retrieving data from: %s", ioc_code, url)
    try:
        df = pd.read_html(url, header=0)[0]
    except ValueError as exc:
        if str(exc) == "No tables found":
            logger.info("%s: No data", ioc_code)
        else:
            logger.exception("%s: Something went wrong", ioc_code)
        raise
    df = normalize_ioc_station_data(ioc_code=ioc_code, df=df, truncate_seconds=truncate_seconds)
    return df


def get_ioc_data(
    ioc_metadata: pd.DataFrame,
    endtime: Union[str, datetime.date] = datetime.date.today(),
    period: float = 1,  # one day
    truncate_seconds: bool = True,
    rate_limit: RateLimit = RateLimit(),
    disable_progress_bar: bool = False,
) -> xr.Dataset:
    """
    Return the data of the stations specified in ``ioc_metadata`` as an ``xr.Dataset``.

    ``truncate_seconds`` needs some explaining. IOC has more than 1000 stations.
    When you retrieve data from all (or at least most of) these stations, you
    end up with thousands of timestamps that only contain a single datapoint.
    This means that the returned ``xr.Dataset`` will contain a huge number of ``NaN``
    which means that you will need a huge amount of RAM.

    In order to reduce the amount of the required RAM we reduce the number of timestamps
    by truncating the seconds. This is how this works:

        2014-01-03 14:53:02 -> 2014-01-03 14:53:00
        2014-01-03 14:53:32 -> 2014-01-03 14:53:00
        2014-01-03 14:53:48 -> 2014-01-03 14:53:00
        2014-01-03 14:54:09 -> 2014-01-03 14:54:00
        2014-01-03 14:54:48 -> 2014-01-03 14:54:00

    Nevertheless this approach has a downside. If a station returns multiple datapoints
    within the same minute, then we end up with duplicate timestamps. When this happens
    we only keep the first datapoint and drop the subsequent ones. So potentially you
    may not retrieve all of the available data.

    If you don't want this behavior, set ``truncate_seconds`` to ``False`` and you
    will retrieve the full data.

    :param ioc_metadata: A ``pd.DataFrame`` returned by ``get_ioc_stations``.
    :param endtime: The date of the "end" of the data.
    :param period: The number of days to be requested. IOC does not support values greater than 30
    :param truncate_seconds: If ``True`` then timestamps are truncated to minutes (seconds are dropped)
    :param rate_limit: The default rate limit is 5 requests/second.
    :param disable_progress_bar: If ``True`` then the progress bar is not displayed.

    """
    if period > IOC_MAX_DAYS_PER_REQUEST:
        msg = (
            f"Unsupported period. Please choose a period smaller than {IOC_MAX_DAYS_PER_REQUEST}: {period}"
        )
        raise ValueError(msg)

    func_kwargs = []
    for ioc_code in ioc_metadata.ioc_code:
        func_kwargs.append(
            dict(
                period=period,
                endtime=endtime,
                ioc_code=ioc_code,
                rate_limit=rate_limit,
                truncate_seconds=truncate_seconds,
            ),
        )

    results = multithread(
        func=get_ioc_station_data,
        func_kwargs=func_kwargs,
        n_workers=5,
        print_exceptions=False,
        disable_progress_bar=disable_progress_bar,
    )

    datasets = []
    for result in results:
        if result.result is not None:
            df = result.result
            meta = ioc_metadata[ioc_metadata.ioc_code == result.kwargs["ioc_code"]]  # type: ignore[index]
            ds = df.set_index(["ioc_code", "time"]).to_xarray()
            ds["lon"] = ("ioc_code", meta.lon)
            ds["lat"] = ("ioc_code", meta.lat)
            ds["country"] = ("ioc_code", meta.country)
            ds["location"] = ("ioc_code", meta.location)
            datasets.append(ds)

    # in order to keep memory consumption low, let's group the datasets
    # and merge them in batches
    while len(datasets) > 5:
        datasets = merge_datasets(datasets)
    # Do the final merging
    ds = xr.merge(datasets)
    return ds



"""
interface with the European Marine Observation and Data Network (EMODnet) API
https://erddap.emodnet-physics.eu/erddap
"""
import datetime
from typing import Optional

import pandas as pd
import pydantic

from searvey.erddap import query_erddap
from searvey.models import ERDDAPDataset
from searvey.models import SymmetricBBox
from searvey.models import SymmetricConstraints

_CRITECH_NORMALIZED_NAMES = {
    "stationID": "station_id",
    "stationName": "station_name",
    "stationSensor": "station_sensor",
    "stationProvider": "station_provider",
    "time (UTC)": "time",
    "latitude (degrees_north)": "lat",
    "longitude (degrees_east)": "lon",
    "SLEV (m)": "sea_level",
    "timestamp (UTC)": "timestamp",
}


EMODNET_CRITECH = ERDDAPDataset(
    server_url=pydantic.parse_obj_as(pydantic.HttpUrl, "https://erddap.emodnet-physics.eu/erddap"),
    dataset_id="TAD_Tsunami_Alert_Device",
    is_longitude_symmetric=True,
)


def normalize_names(df: pd.DataFrame) -> pd.DataFrame:
    df = df.rename(columns=_CRITECH_NORMALIZED_NAMES)
    return df


def remove_null_sea_levels(df: pd.DataFrame) -> pd.DataFrame:
    df = df[~df.sea_level.isna()]
    return df


def normalize_timestamps(df: pd.DataFrame) -> pd.DataFrame:
    df = df.assign(time=pd.to_datetime(df.time))
    return df


def make_categories(df: pd.DataFrame) -> pd.DataFrame:
    df = df.assign(
        station_id=df.station_id.astype("category"),
        station_name=df.station_name.astype("category"),
        station_sensor=df.station_sensor.astype("category"),
        station_provider=df.station_provider.astype("category"),
        author=df.author.astype("category"),
        command=df.command.astype("category"),
    )
    return df


def get_critech_data(
    start_date: datetime.datetime,
    end_date: Optional[datetime.datetime] = None,
    lat_min: float = -90,
    lat_max: float = 90,
    lon_min: float = -180,
    lon_max: float = 180,
    timeout: int = 60,
) -> pd.DataFrame:
    bbox = SymmetricBBox(
        lat_min=lat_min,
        lat_max=lat_max,
        lon_min=lon_min,
        lon_max=lon_max,
    )
    constraints = SymmetricConstraints(
        bbox=bbox,
        start_date=start_date,
        end_date=end_date or datetime.datetime.now(),
    )
    df = query_erddap(dataset=EMODNET_CRITECH, constraints=constraints, timeout=timeout)
    df = normalize_names(df)
    df = remove_null_sea_levels(df)
    df = normalize_timestamps(df)
    df = make_categories(df)
    return df



"""
interface with the University of Hawai'i at Manoa School of Ocean and Earth Science and Technology (SOEST) Sea Level Center (UHSLC) API
https://uhslc.soest.hawaii.edu/erddap
"""
import datetime
from typing import Optional

import pandas as pd
import pydantic

from searvey.erddap import query_erddap
from searvey.models import AsymmetricBBox
from searvey.models import AsymmetricConstraints
from searvey.models import ERDDAPDataset
from searvey.utils import lon3_to_lon1


_UHSLC_NORMALIZED_NAMES = {
    "sea_level (millimeters)": "sea_level",
    "time (UTC)": "time",
    "latitude (degrees_north)": "lat",
    "longitude (degrees_east)": "lon",
    "last_rq_date (UTC)": "last_rq_date",
}

SOEST_UHSLC = ERDDAPDataset(
    server_url=pydantic.parse_obj_as(pydantic.HttpUrl, "https://uhslc.soest.hawaii.edu/erddap"),
    dataset_id="global_hourly_fast",
    is_longitude_symmetric=False,
)


def normalize_names(df: pd.DataFrame) -> pd.DataFrame:
    df = df.rename(columns=_UHSLC_NORMALIZED_NAMES)
    return df


def normalize_longitudes(df: pd.DataFrame) -> pd.DataFrame:
    df = df.assign(lon=lon3_to_lon1(df.lon))
    return df


def remove_null_sea_levels(df: pd.DataFrame) -> pd.DataFrame:
    df = df[~df.sea_level.isna()]
    return df


def normalize_timestamps(df: pd.DataFrame, freq: str = "H") -> pd.DataFrame:
    df = df.assign(time=pd.to_datetime(df.time).dt.round(freq=freq))
    return df


def normalize_sea_level(df: pd.DataFrame) -> pd.DataFrame:
    df = df.assign(sea_level=df.sea_level * 1000)
    return df


def make_categories(df: pd.DataFrame) -> pd.DataFrame:
    df = df.assign(
        # gloss_id=df.gloss_id.astype(int),
        station_name=df.station_name.astype("category"),
        station_country=df.station_country.astype("category"),
        station_country_code=df.station_country_code.astype("category"),
        ssc_id=df.ssc_id.astype("category"),
        uhslc_id=df.uhslc_id.astype("category"),
        gloss_id=df.gloss_id.astype("category"),
        record_id=df.record_id.astype("category"),
    )
    return df


def get_uhslc_data(
    start_date: datetime.datetime,
    end_date: Optional[datetime.datetime] = None,
    lat_min: float = -90,
    lat_max: float = 90,
    lon_min: float = 0,
    lon_max: float = 360,
    timeout: int = 10,
) -> pd.DataFrame:
    bbox = AsymmetricBBox(
        lat_min=lat_min,
        lat_max=lat_max,
        lon_min=lon_min,
        lon_max=lon_max,
    )
    constraints = AsymmetricConstraints(
        bbox=bbox,
        start_date=start_date,
        end_date=end_date or datetime.datetime.now(),
    )
    df = query_erddap(dataset=SOEST_UHSLC, constraints=constraints, timeout=timeout)
    df = normalize_names(df)
    df = normalize_longitudes(df)
    df = remove_null_sea_levels(df)
    df = normalize_timestamps(df)
    df = normalize_sea_level(df)
    df = make_categories(df)
    return df



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['searvey']

package_data = \
{'': ['*']}

install_requires = \
['Shapely',
 'beautifulsoup4',
 'erddapy',
 'geopandas',
 'html5lib',
 'limits',
 'lxml',
 'numpy',
 'pandas',
 'pydantic',
 'requests',
 'tqdm',
 'typepigeon',
 'xarray']

setup_kwargs = {
    'name': 'searvey',
    'version': '0.2.1',
    'description': '',
    'long_description': '# searvey\n\n[![pre-commit.ci](https://results.pre-commit.ci/badge/github/oceanmodeling/searvey/master.svg)](https://results.pre-commit.ci/latest/github/oceanmodeling/searvey/master)\n[![tests](https://github.com/oceanmodeling/searvey/actions/workflows/run_tests.yml/badge.svg)](https://github.com/oceanmodeling/searvey/actions/workflows/run_tests.yml)\n[![readthedocs](https://readthedocs.org/projects/pip/badge/)](https://readthedocs.org/projects/searvey)\n[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/oceanmodeling/searvey/master?urlpath=%2Flab)\n\nSearvey aims to provide the following functionality:\n\n- Unified catalogue of observational data including near real time.\n\n- Real time data analysis/clean up to facilitate comparison with numerical\n  models.\n\n- On demand data retrieval from multiple sources that currently include:\n\n    - U.S. Center for Operational Oceanographic Products and Services (CO-OPS)\n    - Flanders Marine Institute (VLIZ); Intergovernmental Oceanographic Commission (IOC)\n\n## Installation\n\nThe package can be installed with `conda`:\n\n`conda install -c conda-forge searvey`\n\n## Development\n\n```\npython3 -mvenv .venv\nsource .venv/bin/activate\npoetry install\npre-commit install\n```\n\nIf you wish to use jupyterlab to test searvey, then, assuming you have an\nexisting jupyterlab\ninstallation, you should be able to add a kernel to it with:\n\n```bash\npython -m ipykernel install --user --name searvey\n```\n',
    'author': 'Panos Mavrogiorgos',
    'author_email': 'pmav99@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/oceanmodeling/searvey.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Helpers for multi processing/threading
"""
from __future__ import annotations

import logging
import os
from collections.abc import Callable
from concurrent.futures import as_completed
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

import pydantic
import tqdm

# FTR, `loky.ProcessPoolExecutor` is more robust WRT pickling big objects
# But since we are not using MultiProcessing here, we don't need to introduce
# an additional dependency
# from loky import ProcessPoolExecutor

logger = logging.getLogger(__name__)


# https://docs.python.org/3/library/os.html#os.cpu_count
try:
    MAX_AVAILABLE_PROCESSES = len(os.sched_getaffinity(0))
except AttributeError:
    MAX_AVAILABLE_PROCESSES = os.cpu_count()
    if MAX_AVAILABLE_PROCESSES is None:
        MAX_AVAILABLE_PROCESSES = 1


class FutureResult(pydantic.BaseModel):
    exception: Optional[Exception] = None
    kwargs: Optional[Dict[str, Any]] = None
    result: Any = None

    class Config:
        arbitrary_types_allowed: bool = True

    def __hash__(self) -> int:
        return hash((type(self),) + tuple(self.__dict__.values()))


def multi(
    executor: Union[Type[ProcessPoolExecutor], Type[ThreadPoolExecutor]],
    func: Callable[..., Any],
    func_kwargs: List[Dict[str, Any]],
    n_workers: int,
    print_exceptions: bool = True,
    include_kwargs: bool = True,
    initializer: Optional[Callable[..., Any]] = None,
    disable_progress_bar: bool = True,
) -> List[FutureResult]:
    with tqdm.tqdm(total=len(func_kwargs), disable=disable_progress_bar) as progress_bar:
        with executor(max_workers=n_workers, initializer=initializer) as xctr:
            futures_to_kwargs = {xctr.submit(func, **kwargs): kwargs for kwargs in func_kwargs}
            results = []
            for future in as_completed(futures_to_kwargs):
                result_kwargs: Optional[Dict[str, Any]] = futures_to_kwargs[future]
                try:
                    func_result = future.result()
                except Exception as exc:
                    if print_exceptions:
                        print(f"<{result_kwargs}> generated an exception: {exc}")
                    if not include_kwargs:
                        result_kwargs = None
                    results.append(FutureResult(exception=exc, kwargs=result_kwargs))
                else:
                    if not include_kwargs:
                        result_kwargs = None
                    results.append(FutureResult(result=func_result, kwargs=result_kwargs))
                finally:
                    progress_bar.update(1)
            return results


def multithread(
    func: Callable[..., Any],
    func_kwargs: List[Dict[str, Any]],
    n_workers: int = max(1, MAX_AVAILABLE_PROCESSES - 1),
    print_exceptions: bool = True,
    include_kwargs: bool = True,
    executor: Type[ThreadPoolExecutor] = ThreadPoolExecutor,
    initializer: Optional[Callable[..., Any]] = None,
    disable_progress_bar: bool = True,
) -> List[FutureResult]:
    results = multi(
        executor=executor,
        func=func,
        func_kwargs=func_kwargs,
        n_workers=n_workers,
        print_exceptions=print_exceptions,
        include_kwargs=include_kwargs,
        initializer=initializer,
        disable_progress_bar=disable_progress_bar,
    )
    return results


def multiprocess(
    func: Callable[..., Any],
    func_kwargs: List[Dict[str, Any]],
    n_workers: int = max(1, MAX_AVAILABLE_PROCESSES - 1),
    print_exceptions: bool = True,
    include_kwargs: bool = True,
    executor: Type[ProcessPoolExecutor] = ProcessPoolExecutor,
    initializer: Optional[Callable[..., Any]] = None,
    disable_progress_bar: bool = True,
) -> List[FutureResult]:
    if n_workers > MAX_AVAILABLE_PROCESSES:
        msg = f"The maximum available processes are {MAX_AVAILABLE_PROCESSES}, not: {n_workers}"
        raise ValueError(msg)
    results = multi(
        executor=executor,
        func=func,
        func_kwargs=func_kwargs,
        n_workers=n_workers,
        print_exceptions=print_exceptions,
        include_kwargs=include_kwargs,
        initializer=initializer,
        disable_progress_bar=disable_progress_bar,
    )
    return results
